/*
 * Copyright 1991 Addison-Wesley Publishing Company
 *                (see the file "COPYRIGHTS" for details)
 */

#include <stdio.h>

#include <Xm/Xm.h>
#include <Xm/DrawingA.h>

#include "IconBox.h"

/*********************** Record for each icon ************************/

typedef struct {
    int        x, y;            /* left upper edge of icon           */
    int        string_left;     /* how much strings extends to left  */
    int        higher;          /* next higher stacked icon          */
    XmString   label;           /* label string                      */
    String     class;           /* for selection of pixmap           */
    String     pix_name;        /* name of icon bitmap               */
    Pixmap     pix;             /* icon pixmap                       */
    String     mask_name;       /* name of mask bitmap               */
    Pixmap     mask;            /* mask bitmap                       */
    Pixel      foreground;      /* foreground for icon pixmap        */
    Pixel      background;      /* background for icon pixmap        */
} IconRec, *IconRecPtr;

/*********************** Icon box data record ************************/

typedef struct {
    String     pathname;        /* pathname where to read info       */
    int        width, height;   /* size of box                       */
    int        icon_width;      /* size of one icon                  */
    int        icon_height;
    int        string_height;   /* common height of string box       */
    int        icon_spacing;    /* spacing between icons             */
    int        shadow_off;      /* offset of shadow effect           */
    int        columns;         /* no of icon columns initially      */
    IconRec   *icons;           /* icon array                        */
    int        no_icons;        /* elements in array                 */
    int        lowest;          /* index of icon at back of stack    */
    Pixel      shadow_color;    /* color for shadow effect           */
    Pixel      background;      /* pixmap background                 */
    Boolean    exposed;         /* to check for first expose         */
    GC         draw_gc;         /* to draw into buffer               */
    GC         normal_gc;       /* to copy to window and for strings */
    Pixmap     back;            /* background storage                */
    XmFontList fontlist;        /* for strings                       */
    int        x_off, y_off;    /* mouse offset when drag-moving     */
    int        moving;          /* index of currently moved icon     */
} BoxRec, *BoxRecPtr;

/********************* Resource list for icon record *****************/

static XtResource icon_resources [] = {
    { XmNiconBitmap, XmCIconBitmap, XmRString, sizeof (Pixmap),
      XtOffset (IconRecPtr, pix_name), XmRImmediate, (caddr_t) None},
    { XmNiconMask, XmCIconMask, XmRString, sizeof (String),
      XtOffset (IconRecPtr, mask_name), XmRImmediate, (caddr_t) None},
    { XmNiconForeground, XmCForeground, XmRPixel, sizeof (Pixel),
      XtOffset (IconRecPtr, foreground), XmRString, "black"},
    { XmNiconBackground, XmCBackground, XmRPixel, sizeof (Pixel),
      XtOffset (IconRecPtr, background), XmRString, "white"}
};

/******************* Resource list for box record ********************/

static XtResource box_resources[] = {
    { XmNpathname, XmCPathname, XmRString, sizeof (String),
      XtOffset (BoxRecPtr, pathname), XmRString, "dir"},
    { XmNiconWidth, XmCIconWidth, XmRInt, sizeof (int),
      XtOffset (BoxRecPtr, icon_width), XmRImmediate, (caddr_t) 40},
    { XmNiconHeight, XmCIconHeight, XmRInt, sizeof (int),
      XtOffset (BoxRecPtr, icon_height), XmRImmediate, (caddr_t) 40},
    { XmNiconSpacing, XmCIconSpacing, XmRInt, sizeof (int),
      XtOffset (BoxRecPtr, icon_spacing), XmRImmediate, (caddr_t) 10},
    { XmNshadowOffset, XmCShadowOffset, XmRInt, sizeof (int),
      XtOffset (BoxRecPtr, shadow_off), XmRImmediate, (caddr_t) 3},
    { XmNcolumns, XmCColumns, XmRInt, sizeof (int),
      XtOffset (BoxRecPtr, columns), XmRImmediate, (caddr_t) 5},
    { XmNshadowColor, XmCShadowColor, XmRPixel, sizeof (Pixel),
      XtOffset (BoxRecPtr, shadow_color), XmRString, "gray50"},
    { XmNfontList, XmCFontList, XmRFontList, sizeof (XmFontList),
      XtOffset (BoxRecPtr, fontlist), XmRString, "fixed"}
};

/******************** Create box and read icons **********************/

/*------------------------------------*/
/* forward declarations for callbacks */
/*------------------------------------*/

static void ExposeBox ();
static void HandleButtons ();
static void HandleMove ();
static void HandleResize ();
static void DestroyIconBox ();

/*----------------------------------------------*/
/* temporary structure for icons read from file */
/*----------------------------------------------*/

typedef struct {
    char    *n;        /* name of icon                    */
    char    *c;        /* class that may determine pixmap */
    int      x, y;     /* position if not -1              */
} TempRec;

/*----------------------*/
/* init one icon record */
/*----------------------*/

static void InitIcon (widget, temp, rec, index)
    Widget                widget;
    TempRec              *temp;
    BoxRec               *rec;
    int                   index;
{
    IconRec              *icon = &rec->icons[index];
    int                   dummy, i;

    /* get resources for icon record */
    XtGetSubresources (widget, icon, temp->n, temp->c,
                       icon_resources, XtNumber (icon_resources), NULL, 0);

    /* set label and its offset to the left of icon */
    icon->label = XmStringCreateLtoR (temp->n, XmSTRING_DEFAULT_CHARSET);
    icon->string_left = ((int) XmStringWidth (rec->fontlist, icon->label) -
                           rec->icon_width + 1) / 2;
    if (icon->string_left < 0) icon->string_left = 0;
    icon->class = temp->c;

    /* calculate initial position */
    if (temp->x < 0) {
        icon->x = (index % rec->columns) *
                  (rec->icon_width + rec->icon_spacing) +
                   rec->icon_spacing/2;
        icon->y = (index / rec->columns) *
                  (rec->icon_height + rec->string_height + rec->icon_spacing) +
                   rec->icon_spacing/2;
    } else {
        icon->x = temp->x; 
        icon->y = temp->y;
    }

    /* icons are read in stacking order */
    icon->higher = index + 1;

    /* get cached pixmap from Motif */
    icon->pix = XmGetPixmap (XDefaultScreenOfDisplay (XtDisplay (widget)),
                  icon->pix_name, icon->foreground, icon->background);

    /* mask must be bitmap, so get and cache it ourself */
    for (i = 0; i < index; i++ )
        if (strcmp (rec->icons[i].mask_name, icon->mask_name) == 0) {
            icon->mask = rec->icons[i].mask;
            break;
        }
    if (i == index)         /* not found */
        if (XReadBitmapFile (XtDisplay (widget),
                   XDefaultRootWindow (XtDisplay (widget)),
                   icon->mask_name, &dummy, &dummy, &icon->mask,
                   &dummy, &dummy) != BitmapSuccess)
        icon->mask = None;
}
                    
/*---------------------------------------*/
/* read the icon information from a file */
/*---------------------------------------*/

static void ReadIcons (widget, rec)
    Widget                widget;
    BoxRec               *rec;
{
    FILE                 *fp;
    char                  nbuf[100], cbuf[100];
    int                   xbuf, ybuf;
    int                   w = 0, h = 0;
    TempRec               tmp[400];
    int                   no, i;
    
    fp = fopen (rec->pathname, "r");
    if (fp == NULL) return;

    no = 0;
    while (fscanf (fp, "%s", nbuf) != EOF) {
        xbuf = -1; ybuf = -1;
        /* check for window size */
        if (strlen (nbuf) == 1 && nbuf[0] == '#')
            fscanf (fp, "%d %d %s", &w, &h, nbuf);
        /* test if position information has been written */
        if (strlen (nbuf) == 1 && nbuf[0] == ':') 
            fscanf (fp, "%d %d %s %s", &xbuf, &ybuf, nbuf, cbuf);
        else
            fscanf (fp, "%s", cbuf);
        /* copy into tmp array */
        tmp[no].x = xbuf;
        tmp[no].y = ybuf;
        tmp[no].n = XtMalloc (strlen (nbuf) + 1);
        strcpy (tmp[no].n, nbuf);
        tmp[no].c = XtMalloc (strlen (cbuf) + 1);
        strcpy (tmp[no].c, cbuf);
        no++;
    }
    fclose (fp);

    /* allocate icon records and initialize */
    rec->icons = (IconRec*) XtCalloc (no, sizeof (IconRec));
    for (i = 0; i < no; i++) {
         InitIcon (widget, &tmp[i], rec, i);
    }
    rec->no_icons = no;
    rec->lowest = 0;
    rec->icons[no-1].higher = -1;

    /* set initial size of window */
    if (w > 0 && h > 0) {
        rec->width = w;
        rec->height = h;
    } else {
        rec->width = (rec->columns + 1) * 
                   (rec->icon_width + rec->icon_spacing);
        rec->height = (((no - 1)/ rec->columns) + 2) *
                   (rec->icon_height + rec->string_height + rec->icon_spacing);
    }
}

/*---------------------------------*/
/*** Exported creation procedure ***/
/*---------------------------------*/

Widget CreateIconBox (parent, add_args, num_add_args)
    Widget          parent;
    Arg             add_args[];
    Cardinal        num_add_args;
{
    Widget          area;
    Arg             args[3];
    XmString        face;
    BoxRec         *rec;
    XGCValues       gc_values;
    XmString        dummy;
    
    rec = (BoxRec *)XtMalloc (sizeof (BoxRec));

    area = XtCreateManagedWidget (
                 "area",
                 xmDrawingAreaWidgetClass,
                 parent,
                 add_args, num_add_args);

    XtGetSubresources (parent, rec, "area", "XmDrawingArea",
                       box_resources, XtNumber (box_resources),
                       add_args, num_add_args);

    dummy = XmStringCreateLtoR ("hg", XmSTRING_DEFAULT_CHARSET);
    rec->string_height = XmStringHeight (rec->fontlist, dummy) + 2;
    XmStringFree (dummy);

    ReadIcons (area, rec);

    XtSetArg (args[0], XmNwidth, rec->width);
    XtSetArg (args[1], XmNheight, rec->height);
    XtSetArg (args[2], XmNuserData, rec);
    XtSetValues (area, args, 3);

    XtAddCallback (area, XmNexposeCallback, ExposeBox, rec);
    XtAddCallback (area, XmNinputCallback, HandleButtons, rec);
    XtAddEventHandler (area, ButtonMotionMask, False, HandleMove, rec);
    XtAddCallback (area, XmNresizeCallback, HandleResize, rec);
    XtAddCallback (area, XmNdestroyCallback, DestroyIconBox, rec);

    return area;
}

/******************** Handling expose through back pixmap ************/

/*--------------------------------*/
/* redraw one icon to back pixmap */
/*--------------------------------*/

static void RedrawIcon (widget, rec, index)
    Widget                widget;
    BoxRec               *rec;
    int                   index;
{
    IconRec    *icon = &rec->icons[index];

    /* set clip mask in gc */
    XSetClipMask (XtDisplay (widget), rec->draw_gc, icon->mask);

    /* fill shadow */
    XSetClipOrigin (XtDisplay (widget), rec->draw_gc,
                      icon->x + rec->shadow_off, icon->y + rec->shadow_off);
    XSetForeground (XtDisplay (widget), rec->draw_gc, rec->shadow_color); 
    XFillRectangle (XtDisplay (widget), rec->back, rec->draw_gc, 
                      icon->x + rec->shadow_off, icon->y + rec->shadow_off,
                      rec->icon_width, rec->icon_height);

    /* draw icon */
    XSetClipOrigin (XtDisplay (widget), rec->draw_gc, icon->x, icon->y);
    XSetForeground (XtDisplay (widget), rec->draw_gc, icon->foreground);
    XSetBackground (XtDisplay (widget), rec->draw_gc, icon->background);
    if (icon->pix != XmUNSPECIFIED_PIXMAP)
        XCopyArea (XtDisplay (widget), icon->pix, rec->back,
                   rec->draw_gc, 0, 0, rec->icon_width, rec->icon_height,
                   icon->x, icon->y);

    /* draw string */
    XmStringDraw (XtDisplay (widget), rec->back,
               rec->fontlist, icon->label, rec->normal_gc,
               icon->x - icon->string_left, icon->y + rec->icon_height + 1,
               rec->icon_width + 2 * icon->string_left,
               XmALIGNMENT_CENTER, XmSTRING_DIRECTION_L_TO_R, NULL);
}

/*------------------------------------*/
/* create back pixmap on first expose */
/*------------------------------------*/

static void InitializeBoxWindow (widget, rec)
    Widget                widget;
    BoxRec               *rec;
{
    Arg         args[3];
    Cardinal    depth;
    Pixel       window_foreground, window_background;
    XGCValues   gc_values;
    int         i;

    /* get values from drawing area */
    XtSetArg (args[0], XmNdepth, &depth);
    XtSetArg (args[1], XmNbackground, &rec->background);
    XtSetArg (args[2], XmNforeground, &window_foreground);
    XtGetValues (widget, args, 3);

    /* create backing pixmap */
    rec->back = XCreatePixmap (XtDisplay (widget), XtWindow (widget),
                   rec->width, rec->height, depth);
    rec->exposed = True;

    /* create writable GC */
    gc_values.foreground = window_foreground;
    gc_values.background = rec->background;
    gc_values.graphics_exposures = False;
    rec->draw_gc = XCreateGC (XtDisplay (widget), XtWindow (widget),
                     GCForeground | GCBackground | GCGraphicsExposures,
                     &gc_values);

    /* get shareable GC */
    gc_values.font = XLoadFont (XtDisplay (widget), "fixed");
    rec->normal_gc = XtGetGC (widget, 
                 GCForeground | GCBackground | GCFont | GCGraphicsExposures,
                 &gc_values);

    /* initialize pixmap contents */
    XSetForeground (XtDisplay (widget), rec->draw_gc, rec->background);
    XFillRectangle (XtDisplay (widget), rec->back, rec->draw_gc,
                      0, 0, rec->width, rec->height);
    for (i = 0; i < rec->no_icons; i++) 
         RedrawIcon (widget, rec, i);
}

/*---------------------------------------*/
/* test whether two rectangles intersect */
/*---------------------------------------*/

static Boolean Intersect (x1, y1, w1, h1, x2, y2, w2, h2)
    int                   x1, y1, w1, h1, x2, y2, w2, h2;
{
    if (x1 + w1 < x2 || y1 + h1 < y2) return False;
    if (x1 > x2 + w2 || y1 > y2 + h2) return False;
    return True;
}

/*---------------------------------------------------------*/
/* redraw partial region of back pixmap and refresh window */
/*---------------------------------------------------------*/

static void RedrawRegion (widget, rec, x, y, width, height)
    Widget                widget;
    BoxRec               *rec;
    int                   x, y, width, height;
{
    int     i;

    /* erase invalid rectangle in back pixmap */
    XSetForeground (XtDisplay (widget), rec->draw_gc, rec->background);
    XSetClipMask (XtDisplay (widget), rec->draw_gc, None);
    XFillRectangle (XtDisplay (widget), rec->back, rec->draw_gc,
                      x, y, width, height);

    /* redraw affected icons to back pixmap */
    i = rec->lowest;
    do {
        if (Intersect (x, y, width, height,
                       rec->icons[i].x - rec->icons[i].string_left,
                       rec->icons[i].y,
                       rec->icon_width + 2 * rec->icons[i].string_left,
                       rec->icon_height + rec->string_height))
             RedrawIcon (widget, rec, i);
        i = rec->icons[i].higher; }
    while (i >= 0);

    /* refresh window of drawing area */
    XCopyArea (XtDisplay (widget), rec->back, XtWindow (widget),
                rec->normal_gc, x, y, width, height, x, y);
}

/*-------------------------*/
/* simple expose procedure */
/*-------------------------*/

static void ExposeBox (widget, rec, call_data)
    Widget                widget;
    BoxRec               *rec;
    XmDrawingAreaCallbackStruct  *call_data;
{
    XEvent *ev = call_data->event;
    if (ev == NULL) return;

    /* create back pixmap if first time */
    if (!rec->exposed) 
        InitializeBoxWindow (widget, rec);

    /* copy back to fromt */
    XCopyArea (XtDisplay (widget), rec->back, XtWindow (widget),
               rec->normal_gc,
               ev->xexpose.x, ev->xexpose.y,
               ev->xexpose.width, ev->xexpose.height,
               ev->xexpose.x, ev->xexpose.y);
}

/************************ Handle button-down *************************/

/*------------------------------------------*/
/* test whether point in icon or label part */
/*------------------------------------------*/

Boolean PointInIcon (x, y, icon, w, h, sh)
    int                   x;
    int                   y;
    IconRec              *icon;
    int                   w;
    int                   h;
    int                   sh;
{
    /* is in icon part ? */
    if (x >= icon->x && x <= icon->x + w &&
        y >= icon->y && y <= icon->y + h)
        return True;
    /* is in label part ? */
    if (x >= icon->x - icon->string_left &&
        x <= icon->x + w + icon->string_left &&
        y >= icon->y + h &&
        y <= icon->y + h + sh)
        return True;
    /* else false */
    return False;
}

/*----------------------*/
/* which icon was hit ? */
/*----------------------*/

static int Hit (rec, x, y)
    BoxRec               *rec;
    int                   x, y;
{
    int                   i, found = -1;

    i = rec->lowest;
    do {
        if (PointInIcon (x, y, &rec->icons[i], 
                         rec->icon_width, rec->icon_height,
                         rec->string_height))
            found = i;
        i = rec->icons[i].higher;
	}
    while (i >= 0);
    return found;
}

/*------------------*/
/* move icon on top */
/*------------------*/

static void SetOnTop (widget, rec)
    Widget                widget;
    BoxRec               *rec;
{
    int                   topindex = rec->moving;
    int                   i;
    int                   search, previous;

    /* search predecessor */
    for (i = 0; i < rec->no_icons; i++)
        if (rec->icons[i].higher == topindex) 
            previous = i;
        else if (rec->icons[i].higher == -1)
            rec->icons[i].higher = topindex;
    if (rec->lowest == topindex) 
        rec->lowest = rec->icons[topindex].higher;
    else
        rec->icons[previous].higher = rec->icons[topindex].higher;
    rec->icons[topindex].higher = -1;

    /* redraw on top */
    RedrawRegion (widget, rec,
                  rec->icons[topindex].x - rec->icons[topindex].string_left,
                  rec->icons[topindex].y,
                  rec->icon_width + 2 * rec->icons[topindex].string_left,
                  rec->icon_height + rec->string_height);
}

/*-----------------------*/
/* handle input callback */
/*-----------------------*/

static void HandleButtons (widget, rec, call_data)
    Widget                widget;
    BoxRec               *rec;
    XmDrawingAreaCallbackStruct  *call_data;
{
    XEvent   *ev = call_data->event;

    /* activate icon on button press */
    if (ev->xany.type == ButtonPress) {
        rec->moving = Hit (rec, ev->xbutton.x, ev->xbutton.y);
        if (rec->moving < 0) return;
        rec->x_off = ev->xbutton.x - rec->icons[rec->moving].x;
        rec->y_off = ev->xbutton.y - rec->icons[rec->moving].y;
        SetOnTop (widget, rec);
    }
}

/************************* Handle resize callback ********************/

static void HandleResize (widget, rec, call_data)
    Widget                widget;
    BoxRec               *rec;
    XmDrawingAreaCallbackStruct  *call_data;
{
    Arg                   args[2];

    XtSetArg (args[0], XmNwidth, &rec->width);
    XtSetArg (args[1], XmNheight, &rec->height);
    XtGetValues (widget, args, 2);
}

/************************** Handle movement **************************/

/*------------------------------------*/
/* find rectangle invalidated by move */
/*------------------------------------*/

static void FindInvalidRect (rec, new_x, new_y, x, y, width, height)
    BoxRec        *rec;
    int            new_x;
    int            new_y;
    int           *x;
    int           *y;
    int           *width;
    int           *height;
{
    int            xmove, ymove;
    IconRec       *icon = &rec->icons[rec->moving];

    /* check movement in X direction */
    xmove = new_x - icon->x;
    if (xmove > 0) {
        *x = icon->x - icon->string_left;
        *width = rec->icon_width + xmove + 2 * icon->string_left;
    } else {
        *x = icon->x + xmove - icon->string_left;
        *width = rec->icon_width - xmove + 2 * icon->string_left;
    }

    /* check movement in Y direction */
    ymove = new_y - icon->y;
    if (ymove > 0) {
        *y = icon->y;
        *height = rec->icon_height + rec->string_height + ymove;
    } else {
        *y = icon->y + ymove;
        *height = rec->icon_height + rec->string_height - ymove;
    }
}

/*----------------------*/
/* handle motion events */
/*----------------------*/

static void HandleMove (widget, rec, ev)
    Widget         widget;
    BoxRec        *rec;
    XEvent        *ev;
{
    int            new_x, new_y;
    int            x, y, width, height;

    if (rec->moving < 0) return;

    /* constrain motion */
    new_x = ev->xmotion.x - rec->x_off;
    if (new_x < 0)
        new_x = 0;
    if (new_x > rec->width - rec->icon_width)
        new_x = rec->width - rec->icon_width;
    new_y = ev->xmotion.y - rec->y_off;
    if (new_y < 0)
        new_y = 0;
    if (new_y > rec->height - rec->icon_height - rec->string_height)
        new_y = rec->height - rec->icon_height - rec->string_height;

    /* find affected region */
    FindInvalidRect (rec, new_x, new_y, &x, &y, &width, &height);

    /* set new position... */
    rec->icons[rec->moving].x = new_x;
    rec->icons[rec->moving].y = new_y;

    /* ...and redraw region */
    RedrawRegion (widget, rec, x, y, width, height);
}

/************************* Exit and destroy cleanup ******************/

/*------------------------------*/
/* write new state back to file */
/*------------------------------*/

static void WriteIcons (rec)
    BoxRec     *rec;
{
    int         i;
    FILE       *fp;
    String      name;

    fp = fopen (rec->pathname, "w");
    if (fp == NULL) return;

    fprintf (fp, "# %d %d\n", rec->width, rec->height);
    i = rec->lowest;
    do {
        XmStringGetLtoR (rec->icons[i].label, XmSTRING_DEFAULT_CHARSET,
                         &name);
        fprintf (fp, ": %d %d %s %s\n", 
                     rec->icons[i].x, rec->icons[i].y,
                     name, rec->icons[i].class);
        i = rec->icons[i].higher;}
    while (i >= 0);
}
   
/*-----------------------------*/
/*** exported save procedure ***/
/*-----------------------------*/

void SaveIconBox (widget)
    Widget      widget;
{
    Arg         args[1];
    BoxRec     *rec;

    XtSetArg (args[0], XmNuserData, &rec);
    XtGetValues (widget, args, 1);
    WriteIcons (rec);
}

/*-------------------------*/
/* handle destroy callback */
/*-------------------------*/

static void DestroyIconBox (widget, rec, call_data)
    Widget                widget;
    BoxRec               *rec;
    caddr_t               call_data;
{
    int                   i, j;

    /* write new state to file */
    WriteIcons (rec);

    /* free pixmaps */
    for (i = 0; i < rec->no_icons; i++) {
         for (j = 0; j < i; j++)
              if (strcmp (rec->icons[i].mask_name, 
                          rec->icons[j].mask_name) == 0) break;
         if (i == j) 
             XFreePixmap (XtDisplay (widget), rec->icons[i].mask);
    }

    /* free records */
    for (i = 0; i < rec->no_icons; i++) {
        XtFree (rec->icons[i].mask_name);
        XtFree (rec->icons[i].pix_name);
        XmStringFree (rec->icons[i].label);
    }
    XtFree (rec->icons);
    XtFree (rec);
}   
