/*
 * Copyright 1991 Addison-Wesley Publishing Company
 *                (see the file "COPYRIGHTS" for details)
 */

#include <Xm/Xm.h>
#include <Xm/ArrowB.h>
#include <Xm/Label.h>
#include <Xm/DrawingA.h>

#include "StepList.h"

/******************* common information for stepper ******************/

/*-------------------------------*/
/* record associated with widget */
/*-------------------------------*/

typedef struct _stepRec {
    XmStringTable   list;
    int             elements, current;
    Widget          b1, b2, field;
    Dimension       arrow_width;
    XtCallbackProc  call_proc;
    caddr_t         client_data;
} StepRec, *StepRecPtr;;

/*-------------------------*/
/* resource list for items */
/*-------------------------*/

static XtResource item_resources[] = {
    { XmNitems, XmCItems, XmRXmStringTable, sizeof (XmStringTable),
      XtOffset (StepRecPtr, list), XmRImmediate, NULL},
    { XmNitemCount, XmCItemCount, XmRInt, sizeof (int),
      XtOffset (StepRecPtr, elements), XmRImmediate, (caddr_t) 0},
    { XmNarrowWidth, XmCArrowWidth, XmRDimension, sizeof (Dimension),
      XtOffset (StepRecPtr, arrow_width), XmRImmediate, (caddr_t) 16}
};

/*---------------------------------------*/
/* definitions to prefetch fontList info */
/* (before widget is created)            */
/*---------------------------------------*/

typedef struct {
    XmFontList     fontlist;
} ResStruct, *ResPtr;

static XtResource fontlist_resource [] = {
    { XmNfontList, XmCFontList, XmRFontList, sizeof (XmFontList),
      XtOffset (ResPtr, fontlist), XmRString, "fixed"}
};

/************************* Stepper callbacks *************************/

/*---------------------------*/
/* call application callback */
/*---------------------------*/

static void CallCallback (rec, event, previous)
    StepRec       *rec;
    XEvent        *event;
    int            previous;
{
    StepListCallbackStruct cb;

    if (rec->call_proc != NULL) {
        cb.reason = XmCR_VALUE_CHANGED;
        cb.event = event;
        cb.list = rec->list;
        cb.elements = rec->elements;
        cb.current = rec->current;
        cb.previous = previous;
        (* rec->call_proc) (rec->field, rec->client_data, &cb);
    }
}

/*-----------------*/
/* StepUp callback */
/*-----------------*/

static void StepUp (widget, rec, call_data)
    Widget                widget;
    StepRec              *rec;
    XmAnyCallbackStruct  *call_data;
{
    Arg             args[1];
    int             previous;

    if (rec->current > 1) {
        previous = rec->current;
        rec->current--;
        XtSetArg (args[0], XmNlabelString, rec->list[rec->current-1]);
        XtSetValues (rec->field, args, 1);
        XtSetSensitive (rec->b1, rec->current > 1);
        XtSetSensitive (rec->b2, True);
        CallCallback (rec, call_data->event, previous);
    }
}

/*-------------------*/
/* StepDown callback */
/*-------------------*/

static void StepDown (widget, rec, call_data)
    Widget                widget;
    StepRec              *rec;
    XmAnyCallbackStruct  *call_data;
{
    Arg             args[1];
    int             previous;

    if (rec->current < rec->elements) {
        previous = rec->current;
        rec->current++;
        XtSetArg (args[0], XmNlabelString, rec->list[rec->current-1]);
        XtSetValues (rec->field, args, 1);
        XtSetSensitive (rec->b1, True);
        XtSetSensitive (rec->b2, rec->current < rec->elements);
        CallCallback (rec, call_data->event, previous);
    }
}

/*-----------------*/
/* resize callback */
/*-----------------*/

static void Resize (widget, rec, call_data)
    Widget                widget;
    StepRec              *rec;
    XmAnyCallbackStruct  *call_data;
{
    Arg             args[3];
    Dimension       width, height, text_height, arrow_width;
    Position        new_y;

    XtSetArg (args[0], XmNwidth, &width);
    XtSetArg (args[1], XmNheight, &height);
    XtGetValues (widget, args, 2);

    if (width < 3 * rec->arrow_width)
       arrow_width = width/3;
    else
       arrow_width = rec->arrow_width;

    XtSetArg (args[0], XmNheight, &text_height);
    XtGetValues (rec->field, args, 1);

    if (height > text_height)
        new_y = (height - text_height)/2;
    else
        new_y = 0;

    /* stop resizing the drawing area now */
    XtSetArg (args[0], XmNresizePolicy, XmRESIZE_NONE);
    XtSetValues (widget, args, 1);

    /* resize the children */
    XtSetArg (args[0], XmNy, new_y);
    XtSetArg (args[1], XmNwidth, arrow_width);
    XtSetValues (rec->b1, args, 2);

    XtSetArg (args[2], XmNx, arrow_width);
    XtSetValues (rec->b2, args, 3);

    XtSetArg (args[1], XmNwidth, width - 2 * arrow_width);
    XtSetArg (args[2], XmNx, 2 * arrow_width);
    XtSetValues (rec->field, args, 3);
}

/************************ Creation procedure *************************/

/*---------------------------------------*/
/* find the longest XmString of the list */
/*---------------------------------------*/

static XmString Longest (xmlist, no_el, fl)
    XmString        *xmlist;
    int             no_el;
    XmFontList      fl;
{
    int             i, max = 0, maxno = -1, new;

    for (i=0; i<no_el; i++) { 
        if (max < (new = XmStringWidth (fl, xmlist[i]))) {
            max = new; maxno = i;
        }
    }
    if (maxno >= 0)
       return (xmlist[maxno]);
    else
       return NULL;
}

/*-------------------*/
/* prefetch fontList */
/*-------------------*/

static XmFontList GetFontList (parent, name, class)
    Widget     parent;
    char      *name;
    char      *class;
{
    ResStruct  res;

    XtGetSubresources (parent, &res, name, class,
                       fontlist_resource, XtNumber (fontlist_resource),
                       NULL, 0);
    return (res.fontlist);
}

/*------------------------*/
/*** creation procedure ***/
/*------------------------*/

Widget CreateStepList (parent, name, add_args, num_args, proc, closure)
    Widget          parent;
    String          name;
    ArgList         add_args;
    Cardinal        num_args;
    XtCallbackProc  proc;
    caddr_t         closure;
{
    Widget          form, text, ab1, ab2;
    Arg             args[5];
    StepRec         *rec;
    Dimension       text_height;
    XmFontList      text_font;
    ArgList         merge_args;

    /* init callback record */
    rec = (StepRec*) XtMalloc (sizeof (StepRec));
    rec->current = 1;
    rec->call_proc = proc;
    rec->client_data = closure;

    /* drawing area as manager */
    XtSetArg (args[0], XmNmarginWidth, 0);
    XtSetArg (args[1], XmNmarginHeight, 0);
    form = XtCreateManagedWidget ("form", xmDrawingAreaWidgetClass, parent,
                                 args, 2);
    XtAddCallback (form, XmNresizeCallback, Resize, rec);

    /* initialize the rest of the callback record */
    XtGetSubresources (form, rec, name, STEPPER_CLASS,
                       item_resources, XtNumber (item_resources),
                       add_args, num_args);
 
    /* first scroll button */
    XtSetArg (args[0], XmNarrowDirection, XmARROW_UP);
    XtSetArg (args[1], XmNwidth, rec->arrow_width);
    ab1 = XtCreateManagedWidget ("ab1", xmArrowButtonWidgetClass, form,
                                 args, 2);
    XtSetSensitive (ab1, False);
    XtAddCallback (ab1, XmNactivateCallback, StepUp, rec);
    rec->b1 = ab1;

    /* second scroll button */
    XtSetArg (args[0], XmNarrowDirection, XmARROW_DOWN);
    XtSetArg (args[1], XmNwidth, rec->arrow_width);
    XtSetArg (args[2], XmNx, rec->arrow_width);
    ab2 = XtCreateManagedWidget ("ab2", xmArrowButtonWidgetClass, form,
                                 args, 3);
    XtSetSensitive (ab2, rec->elements > 1);
    XtAddCallback (ab2, XmNactivateCallback, StepDown, rec);
    rec->b2 = ab2;

    /* get the fontlist which will be used in the text display */
    text_font = GetFontList (form, name, "XmLabel");

    /* text display (with the longest item) */  
    XtSetArg (args[0], XmNlabelString, 
              Longest (rec->list, rec->elements, text_font));
    XtSetArg (args[1], XmNuserData, rec);
    XtSetArg (args[2], XmNrecomputeSize, False);
    XtSetArg (args[3], XmNx, 2 * rec->arrow_width);
    XtSetArg (args[4], XmNpushButtonEnabled, False);

    merge_args = XtMergeArgLists (add_args, num_args, args, 5);
    text = XtCreateManagedWidget (
                 name,
                 xmLabelWidgetClass,
                 form,
                 merge_args,
                 num_args + 5);
    XtFree (merge_args);
    rec->field = text;

    /* align text and arrows */
    XtSetArg (args[0], XmNheight, &text_height);
    XtGetValues (text, args, 1);
    XtSetArg (args[0], XmNheight, text_height);
    XtSetValues (ab1, args, 1);
    XtSetValues (ab2, args, 1);

    /* set text to first label */
    if (rec->elements > 1) {
        XtSetArg (args[0], XmNlabelString, rec->list[0]);
        XtSetValues (text, args, 1);
    }
}

/*********************** Access current item *************************/

/*-------------------------*/
/*** change current item ***/
/*-------------------------*/

void StepListSetCurrent (widget, new, call)
    Widget        widget;
    int           new;
    Boolean       call;
{
    StepRec      *rec;
    Arg           args[1];
    int           previous;

    XtSetArg (args[0], XmNuserData, &rec);
    XtGetValues (widget, args, 1);

    if (rec == NULL) return;

    if (new < 1 || new > rec->elements) return;

    previous = rec->current;
    rec->current = new;
    XtSetArg (args[0], XmNlabelString, rec->list[rec->current-1]);
    XtSetValues (rec->field, args, 1);
    XtSetSensitive (rec->b1, rec->current > 1);
    XtSetSensitive (rec->b2, rec->current < rec->elements);
    if (call) CallCallback (rec->field, NULL, previous);
}  

/*----------------------*/
/*** get current item ***/
/*----------------------*/

void StepListGetCurrent (widget, current_return, label_return)
    Widget         widget;
    int           *current_return;
    XmString      *label_return;
{
    StepRec       *rec;
    Arg            args[1];

    XtSetArg (args[0], XmNuserData, &rec);
    XtGetValues (widget, args, 1);

    *current_return = 0;    /* invalid data */
    if (rec == NULL) return;

    if (rec->current < 1 || rec->current > rec->elements) return;

    *current_return = rec->current;
    *label_return = rec->list[rec->current-1];
}




